SystemOrganization addCategory: #MUI!

BorderedMorph subclass: #MUIPolygonMorph
	instanceVariableNames: 'vertices closed filledForm arrows arrowForms smoothCurve curveState borderDashSpec handles borderForm'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MUI'!

!MUIPolygonMorph class methodsFor: 'instance creation' stamp: 'yo 9/2/2016 15:27'!
vertices: verts color: c borderWidth: bw borderColor: bc
	^ self basicNew beStraightSegments vertices: verts color: c borderWidth: bw borderColor: bc! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
addCustomMenuItems: aMenu hand: aHandMorph 
	| |
	super addCustomMenuItems: aMenu hand: aHandMorph.
	aMenu
		addUpdating: #handlesShowingPhrase
		target: self
		action: #showOrHideHandles.
	vertices size > 2
		ifTrue: [ self addPolyLIneCurveMenuItems: aMenu hand: aHandMorph ].
	aMenu add: 'specify dashed line' translated action: #specifyDashedLine.
	"aMenu add: 'use debug border' translated action: #showSegmentsBorderStyle."
	self isOpen
		ifTrue: [self addPolyArrowMenuItems: aMenu hand: aHandMorph]
			ifFalse: [self addPolyShapingMenuItems: aMenu hand: aHandMorph]! !

!MUIPolygonMorph methodsFor: 'rotate scale and flex' stamp: 'yo 9/2/2016 15:27'!
addFlexShellIfNecessary
	"When scaling or rotating from a halo, I can do this without a flex shell"

	^ self
! !

!MUIPolygonMorph methodsFor: 'editing' stamp: 'yo 9/6/2016 13:16'!
addHandles
	"Put moving handles at the vertices. Put adding handles at
	edge midpoints.
	Moving over adjacent vertex and dropping will delete a
	vertex. "
	self removeHandles.
	handles := OrderedCollection new.
	vertices
		withIndexDo: [:vertPt :vertIndex | 
			| handle |
			handle := EllipseMorph
						newBounds: (Rectangle center: vertPt extent: 8 @ 8)
						color: (self handleColorAt: vertIndex) .
			self addMorph: handle.
			handles addLast: handle].
	self isCurvy
		ifTrue: [self updateHandles; layoutChanged].
	self changed! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
addPolyArrowMenuItems: aMenu hand: aHandMorph 
aMenu addLine.
			aMenu
				addWithLabel: '---'
				enablement: [self isOpen
						and: [arrows ~~ #none]]
				action: #makeNoArrows.
			aMenu
				addWithLabel: '-->'
				enablement: [self isOpen
						and: [arrows ~~ #forward]]
				action: #makeForwardArrow.
			aMenu
				addWithLabel: '<--'
				enablement: [self isOpen
						and: [arrows ~~ #back]]
				action: #makeBackArrow.
			aMenu
				addWithLabel: '<->'
				enablement: [self isOpen
						and: [arrows ~~ #both]]
				action: #makeBothArrows.
			aMenu add: 'customize arrows' translated action: #customizeArrows:.
			(self hasProperty: #arrowSpec)
				ifTrue: [aMenu add: 'standard arrows' translated action: #standardArrows]! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
addPolyLIneCurveMenuItems: aMenu hand: aHandMorph 

	aMenu addLine;
				addUpdating: #openOrClosePhrase
				target: self
				action: #makeOpenOrClosed.
			
			aMenu
				addUpdating: #smoothOrSegmentedPhrase
				target: self
				action: #toggleSmoothing.! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
addPolyShapingMenuItems: aMenu hand: aHandMorph 
	aMenu addLine.
			aMenu
				addWithLabel: 'make inscribed diamondOval'
				enablement: [self isClosed ]
				action: #diamondOval.
			aMenu
				addWithLabel: 'make enclosing rectangleOval'
				enablement: [self isClosed ]
					action: #rectOval.
					! !

!MUIPolygonMorph methodsFor: 'drawing' stamp: 'yo 9/2/2016 15:27'!
areasRemainingToFill: aRectangle
	"Could be improved by quick check of inner rectangle"

	^ Array with: aRectangle! !

!MUIPolygonMorph methodsFor: 'private' stamp: 'yo 9/2/2016 15:27'!
arrowBoundsAt: endPoint from: priorPoint 
	"Answer a triangle oriented along the line from priorPoint to endPoint."
	| d v angle wingBase arrowSpec length width |
	v := endPoint - priorPoint.
	angle := v degrees.
	d := borderWidth max: 1.
	arrowSpec := self valueOfProperty: #arrowSpec ifAbsent: [5@4].
	length := arrowSpec x abs.  width := arrowSpec y abs.
	wingBase := endPoint + (Point r: d * length degrees: angle + 180.0).
	arrowSpec x >= 0
		ifTrue: [^ {	endPoint.
					wingBase + (Point r: d * width degrees: angle + 125.0).
					wingBase + (Point r: d * width degrees: angle - 125.0) }]
		ifFalse: ["Negative length means concave base."
				^ {	endPoint.
					wingBase + (Point r: d * width degrees: angle + 125.0).
					wingBase.
					wingBase + (Point r: d * width degrees: angle - 125.0) }]! !

!MUIPolygonMorph methodsFor: 'private' stamp: 'yo 9/2/2016 15:27'!
arrowForms
	"ArrowForms are computed only upon demand"
	arrowForms
		ifNotNil: [^ arrowForms].
	arrowForms := Array new.
	self hasArrows
		ifFalse: [^ arrowForms].
	(arrows == #forward
			or: [arrows == #both])
		ifTrue: [arrowForms := arrowForms
						copyWith: (self computeArrowFormAt: vertices last from: self nextToLastPoint)].
	(arrows == #back
			or: [arrows == #both])
		ifTrue: [arrowForms := arrowForms
						copyWith: (self computeArrowFormAt: vertices first from: self nextToFirstPoint)].
	^ arrowForms! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
arrowLength: aLength
	"Assumes that I have exactly two vertices"

	| theta horizontalOffset verticalOffset newTip delta |
	delta := vertices second - vertices first.
	theta := delta theta.
	horizontalOffset := aLength * (theta cos).
	verticalOffset := aLength * (theta sin).
	newTip := vertices first + (horizontalOffset @ verticalOffset).
	self verticesAt: 2 put: newTip! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
arrowSpec: specPt
	"Specify a custom arrow for this line.
	specPt x abs gives the length of the arrow (point to base) in terms of borderWidth.
	If specPt x is negative, then the base of the arrow will be concave.
	specPt y abs gives the width of the arrow.
	The standard arrow is equivalent to arrowSpec: 5@4.
	See arrowBoundsAt:From: for details."

	self setProperty: #arrowSpec toValue: specPt.
	self computeBounds! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
arrows
	^arrows! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
arrowsContainPoint: aPoint
	"Answer an Array of two Booleans that indicate whether the given point is inside either arrow"

	| retval f |

	retval := { false . false }.
	(super containsPoint: aPoint) ifFalse: [^ retval ].
	(closed or: [arrows == #none or: [vertices size < 2]]) ifTrue: [^ retval].

	(arrows == #forward or: [arrows == #both]) ifTrue: [	"arrowForms first has end form"
		f := self arrowForms first.
		retval at: 2 put: ((f pixelValueAt: aPoint - f offset) > 0)
	].
	(arrows == #back or: [arrows == #both]) ifTrue: [ "arrowForms last has start form"
		f := self arrowForms last.
		retval at: 1 put: ((f pixelValueAt: aPoint - f offset) > 0)
	].
	^retval.! !

!MUIPolygonMorph methodsFor: 'initialization' stamp: 'yo 9/2/2016 15:27'!
beSmoothCurve

	smoothCurve == true ifFalse:
		[smoothCurve := true.
		self computeBounds]! !

!MUIPolygonMorph methodsFor: 'initialization' stamp: 'yo 9/2/2016 15:27'!
beStraightSegments

	smoothCurve == false ifFalse:
		[smoothCurve := false.
		self computeBounds]! !

!MUIPolygonMorph methodsFor: 'access' stamp: 'yo 9/2/2016 15:27'!
borderColor: aColor 

	super borderColor: aColor.
	(borderColor isColor and: [borderColor isTranslucentColor]) 
		== (aColor isColor and: [aColor isTranslucentColor]) 
			ifFalse: 
				["Need to recompute fillForm and borderForm
					if translucency of border changes."

				self releaseCachedState]! !

!MUIPolygonMorph methodsFor: 'dashes' stamp: 'yo 9/2/2016 15:27'!
borderDashOffset
	borderDashSpec size < 4 ifTrue: [^0.0].
	^(borderDashSpec fourth) asFloat! !

!MUIPolygonMorph methodsFor: 'private' stamp: 'yo 9/2/2016 15:27'!
borderForm
	"A form must be created for drawing the border whenever the borderColor is translucent."

	| borderCanvas |
	borderForm ifNotNil: [^ borderForm].
	borderCanvas := (Display defaultCanvasClass extent: bounds extent depth: 1)
		shadowColor: Color black.
	borderCanvas translateBy: bounds topLeft negated
		during:[:tempCanvas| self drawBorderOn: tempCanvas].
	borderForm := borderCanvas form.
	self arrowForms do:
		[:f |  "Eliminate overlap between line and arrowheads if transparent."
		borderForm copy: f boundingBox from: f to: f offset - self position rule: Form erase].
	^ borderForm! !

!MUIPolygonMorph methodsFor: 'accessing' stamp: 'yo 9/2/2016 15:27'!
borderWidth: anInteger

	borderColor ifNil: [borderColor := Color black].
	borderWidth := anInteger max: 0.
	self computeBounds! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
bounds: newBounds
	"This method has to be reimplemented since self extent: will also change self bounds origin,
	super bounds would leave me in wrong position when container is growing.
	Always change extent first then position"
	
	self extent: newBounds extent; position: newBounds topLeft
! !

!MUIPolygonMorph methodsFor: 'attachments' stamp: 'yo 9/2/2016 15:27'!
boundsSignatureHash
	^(vertices - (self positionInWorld))  hash
! !

!MUIPolygonMorph methodsFor: 'visual properties' stamp: 'yo 9/2/2016 15:27'!
canHaveFillStyles
	"Return true if the receiver can have general fill styles; not just colors.
	This method is for gradually converting old morphs."
	^true! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
closestPointTo: aPoint 
	| closestPoint minDist |
	closestPoint := minDist := nil.
	self lineSegmentsDo: 
			[:p1 :p2 | | dist curvePoint | 
			curvePoint := aPoint nearestPointOnLineFrom: p1 to: p2.
			dist := curvePoint dist: aPoint.
			(closestPoint isNil or: [dist < minDist]) 
				ifTrue: 
					[closestPoint := curvePoint.
					minDist := dist]].
	^closestPoint! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
closestSegmentTo: aPoint
	"Answer the starting index of my (big) segment nearest to aPoint"
	| closestPoint minDist vertexIndex closestVertexIndex |
	vertexIndex := 0.
	closestVertexIndex := 0.
	closestPoint := minDist := nil.
	self lineSegmentsDo:
		[:p1 :p2 | | dist curvePoint | 
		(p1 = (self vertices at: vertexIndex + 1))
			ifTrue: [ vertexIndex := vertexIndex + 1 ].
		curvePoint := aPoint nearestPointOnLineFrom: p1 to: p2.
		dist := curvePoint dist: aPoint.
		(closestPoint == nil or: [dist < minDist])
			ifTrue: [closestPoint := curvePoint.
					minDist := dist.
					closestVertexIndex := vertexIndex. ]].
	^ closestVertexIndex! !

!MUIPolygonMorph methodsFor: 'smoothing' stamp: 'yo 9/2/2016 15:27'!
coefficients
	"Compute an array for the coefficients."
	| verts vertXs vertYs slopeXs slopeYs coefficients |
	curveState
		ifNotNil: [^ curveState at: 1].
	verts := self vertices.
	verts size < 1
		ifTrue: [^ self].
	"Less than three points handled as segments by our 
	lineSegmentsDo:"
	(self isCurvier)
		ifFalse: [closed
				ifTrue: [verts := verts , verts first asOrderedCollection]].
	coefficients := {
		vertXs := verts collect: [:p | p x asFloat].
		slopeXs := self slopes: vertXs.
		vertXs changeInSlopes: slopeXs.
		vertXs changeOfChangesInSlopes: slopeXs.
		vertYs := verts collect: [:p | p y asFloat].
		slopeYs := self slopes: vertYs.
		vertYs changeInSlopes: slopeYs.
		vertYs changeOfChangesInSlopes: slopeYs.
		Array new: verts size withAll: 12}.
	coefficients
		at: 9
		put: ((1 to: verts size)
				collect: [:i | (coefficients cubicPointPolynomialAt: i) bestSegments]).
	(self isCurvier)
		ifFalse: [closed
				ifTrue: [coefficients := coefficients
								collect: [:each | each allButLast]]].
	curveState := {coefficients. nil. nil}.
	self computeNextToEndPoints.
	^ coefficients! !

!MUIPolygonMorph methodsFor: 'private' stamp: 'yo 9/2/2016 15:27'!
computeArrowFormAt: endPoint from: priorPoint 
	"Compute a triangle oriented along the line from priorPoint to  
	endPoint. Then draw those lines in a form and return that  
	form, with appropriate offset"

	| p1 pts box arrowForm bb origin |
	pts := self arrowBoundsAt: endPoint from: priorPoint.
	box := ((pts first rect: pts last) encompass: (pts second)) expandBy: 1.
	arrowForm := Form extent: box extent asIntegerPoint.
	bb := (BitBlt toForm: arrowForm)
				sourceForm: nil;
				fillColor: Color black;
				combinationRule: Form over;
				width: 1;
				height: 1.
	origin := box topLeft.
	p1 := pts last - origin.
	pts do: 
			[:p | 
			bb drawFrom: p1 to: p - origin.
			p1 := p - origin].
	arrowForm convexShapeFill: Color black.
	^arrowForm offset: box topLeft! !

!MUIPolygonMorph methodsFor: 'private' stamp: 'yo 9/2/2016 15:27'!
computeBounds
	| oldBounds delta excludeHandles |
	vertices ifNil: [^ self].

	self changed.
	oldBounds := bounds.
	self releaseCachedState.
	bounds := self curveBounds expanded copy.
	self arrowForms do:
		[:f | bounds swallow: (f offset extent: f extent)].
	handles ifNotNil: [self updateHandles].

	"since we are directly updating bounds, see if any ordinary submorphs exist and move them accordingly"
	(oldBounds notNil and: [(delta := bounds origin - oldBounds origin) ~= (0@0)]) ifTrue: [
		excludeHandles := IdentitySet new.
		handles ifNotNil: [excludeHandles addAll: handles].
		self submorphsDo: [ :each |
			(excludeHandles includes: each) ifFalse: [
				each position: each position + delta
			].
		].
	].
	self layoutChanged.
	self changed.
! !

!MUIPolygonMorph methodsFor: 'smoothing' stamp: 'yo 9/2/2016 15:27'!
computeNextToEndPoints
	| pointAfterFirst pointBeforeLast |
	pointAfterFirst := nil.
	self lineSegmentsDo: 
			[:p1 :p2 | 
			pointAfterFirst ifNil: [pointAfterFirst := p2 asIntegerPoint].
			pointBeforeLast := p1 asIntegerPoint].
	curveState at: 2 put: pointAfterFirst.
	curveState at: 3 put: pointBeforeLast! !

!MUIPolygonMorph methodsFor: 'geometry testing' stamp: 'yo 9/2/2016 15:27'!
containsPoint: aPoint
	(super containsPoint: aPoint) ifFalse: [^ false].

	closed & color isTransparent not ifTrue:
		[^ (self filledForm pixelValueAt: aPoint - bounds topLeft + 1) > 0].

	self lineSegmentsDo:
		[:p1 :p2 |
		(aPoint onLineFrom: p1 to: p2 within: (3 max: borderWidth+1//2) asFloat)
				ifTrue: [^ true]].

	self arrowForms do:
		[:f | (f pixelValueAt: aPoint - f offset) > 0 ifTrue: [^ true]].

	^ false! !

!MUIPolygonMorph methodsFor: 'rounding' stamp: 'yo 9/2/2016 15:27'!
cornerStyle: aSymbol 
	"Set the receiver's corner style.  But, in this case, do *not*"

	self removeProperty: #cornerStyle.
	self changed! !

!MUIPolygonMorph methodsFor: 'accessing' stamp: 'yo 9/2/2016 15:27'!
couldHaveRoundedCorners
	^ false! !

!MUIPolygonMorph methodsFor: 'private' stamp: 'yo 9/2/2016 15:27'!
curveBounds
	"Compute the bounds from actual curve traversal, with 
	leeway for borderWidth. 
	Also note the next-to-first and next-to-last points for arrow 
	directions."
	"wiz - to avoid roundoff errors we return unrounded curvebounds."
	"we expect our receiver to take responsibility for approriate rounding adjustment."
	"hint: this is most likely 'self curveBounds expanded' "
	| pointAfterFirst pointBeforeLast  oX oY cX cY |
	self isCurvy
		ifFalse: [^ (Rectangle encompassing: vertices)
				expandBy: borderWidth * 0.5 ].
	curveState := nil.
	"Force recomputation"
	"curveBounds := vertices first corner: vertices last."
	pointAfterFirst := nil.
	self
		lineSegmentsDo: [:p1 :p2 | 
			pointAfterFirst isNil
				ifTrue: [pointAfterFirst := p2 floor .
					oX := cX := p1 x.
					oY := cY := p1 y. ].
			"curveBounds := curveBounds encompass: p2 ."
			oX:= oX min: p2 x.
			cX := cX max: p2 x.
			oY := oY min: p2 y.
			cY := cY max: p2 y.
			pointBeforeLast := p1 floor ].
	curveState at: 2 put: pointAfterFirst.
	curveState at: 3 put: pointBeforeLast.
	^ ( oX @ oY corner: cX @ cY )  expandBy: borderWidth * 0.5 ! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
customizeArrows: evt
	| handle origin aHand |
	aHand := evt ifNil: [self primaryHand] ifNotNil: [evt hand].
	origin := aHand position.
	(handle := HandleMorph new)
		forEachPointDo:
			[:newPoint | handle removeAllMorphs.
			handle addMorph:
				(LineMorph from: origin to: newPoint color: Color black width: 1).
			self arrowSpec: (newPoint - origin) / 5.0]
		lastPointDo:
			[:newPoint | handle deleteBalloon.
			self halo ifNotNil: [:halo | halo addHandles].].
	aHand attachMorph: handle.
	handle setProperty: #helpAtCenter toValue: true.
	handle showBalloon:
'Move cursor left and right
to change arrow length and style.
Move it up and down to change width.
Click when done.' hand: evt hand.
	handle startStepping! !

!MUIPolygonMorph methodsFor: 'dashes' stamp: 'yo 9/2/2016 15:27'!
dashedBorder
	^borderDashSpec
	"A dash spec is a 3- or 5-element array with
		{ length of normal border color.
		length of alternate border color.
		alternate border color.
		starting offset.
		amount to add to offset at each step }
	Starting offset is usually = 0, but changing it moves the dashes along the curve."
! !

!MUIPolygonMorph methodsFor: 'dashes' stamp: 'yo 9/2/2016 15:27'!
dashedBorder: dashSpec
	"A dash spec is a 3- or 5-element array with
		{ length of normal border color.
		length of alternate border color.
		alternate border color.
		starting offset.
		amount to add to offset at each step }
	Starting offset is usually = 0, but changing it moves the dashes along the curve."

	borderDashSpec := dashSpec.
	self changed! !

!MUIPolygonMorph methodsFor: 'attachments' stamp: 'yo 9/2/2016 15:27'!
defaultAttachmentPointSpecs
	^{ 
		{ #firstVertex } .
		{ #midpoint  } .
		{ #lastVertex }
	}! !

!MUIPolygonMorph methodsFor: 'initialization' stamp: 'yo 9/2/2016 15:27'!
defaultBorderColor
	"answer the default border color/fill style for the receiver"
	^ Color
		r: 0.0
		g: 0.419
		b: 0.935! !

!MUIPolygonMorph methodsFor: 'initialization' stamp: 'yo 9/2/2016 15:27'!
defaultColor
	"answer the default color/fill style for the receiver"
	^ Color orange! !

!MUIPolygonMorph methodsFor: 'editing' stamp: 'yo 9/2/2016 15:27'!
deleteVertexAt: anIndex
	"This acts as a backstop for MixedCurveMorph."
			self
				setVertices: (vertices
						copyReplaceFrom: anIndex
						to: anIndex
						with: Array new).
						! !

!MUIPolygonMorph methodsFor: 'smoothing' stamp: 'yo 9/2/2016 15:27'!
derivs: a first: point1 second: point2 third: point3 
	"Compute the first, second and third derivitives (in coeffs) from
	the Points in this Path (coeffs at: 1 and coeffs at: 5)."

	| len v anArray |
	len := a size.
	len < 2 ifTrue: [^self].
	len > 2 
		ifTrue: 
			[v := Array new: len.
			v at: 1 put: 4.0.
			anArray := Array new: len.
			anArray at: 1 put: 6.0 * (a first - (a second * 2.0) + (a third)).
			2 to: len - 2
				do: 
					[:i | 
					v at: i put: 4.0 - (1.0 / (v at: i - 1)).
					anArray at: i
						put: 6.0 * ((a at: i) - ((a at: i + 1) * 2.0) + (a at: i + 2)) 
								- ((anArray at: i - 1) / (v at: i - 1))].
			point2 at: len - 1 put: (anArray at: len - 2) / (v at: len - 2).
			len - 2 to: 2
				by: 0 - 1
				do: 
					[:i | 
					point2 at: i
						put: ((anArray at: i - 1) - (point2 at: i + 1)) / (v at: i - 1)]].
	point2 at: 1 put: (point2 at: len put: 0.0).
	1 to: len - 1
		do: 
			[:i | 
			point1 at: i
				put: (a at: i + 1) - (a at: i) 
						- (((point2 at: i) * 2.0 + (point2 at: i + 1)) / 6.0).
			point3 at: i put: (point2 at: i + 1) - (point2 at: i)]! !

!MUIPolygonMorph methodsFor: 'shaping' stamp: 'yo 9/2/2016 15:27'!
diamondOval
	"Set my vertices to an array of edge midpoint vertices. 
	Order of vertices is in the tradion of warpblt quads."
	| b r |
	b := self bounds.
	r := {b leftCenter. b bottomCenter. b rightCenter. b topCenter}.
	self setVertices: r! !

!MUIPolygonMorph methodsFor: 'drawing' stamp: 'yo 9/2/2016 15:27'!
drawArrowOn: aCanvas at: endPoint from: priorPoint 
	"Draw a triangle oriented along the line from priorPoint to  
	endPoint. Answer the wingBase."

	| pts spec wingBase |
	pts := self arrowBoundsAt: endPoint from: priorPoint.
	wingBase := pts size = 4 
				ifTrue: [pts third]
				ifFalse: [(pts copyFrom: 2 to: 3) average].
	spec := self valueOfProperty: #arrowSpec ifAbsent: [5 @ 4].
	spec x sign = spec y sign 
		ifTrue: [aCanvas drawPolygon: pts fillStyle: borderColor]
		ifFalse: 
			[aCanvas 
				drawPolygon: pts
				fillStyle: Color transparent
				borderWidth: (borderWidth + 1) // 2
				borderColor: borderColor].
	^wingBase! !

!MUIPolygonMorph methodsFor: 'drawing' stamp: 'yo 9/2/2016 15:27'!
drawArrowsOn: aCanvas 
	"Answer (possibly modified) endpoints for border drawing"
	"ArrowForms are computed only upon demand"
	| array |

	self hasArrows
		ifFalse: [^ #() ].
	"Nothing to do"

	array := Array with: vertices first with: vertices last.

	"Prevent crashes for #raised or #inset borders"
	borderColor isColor
		ifFalse: [ ^array ].

	(arrows == #forward or: [arrows == #both])
		ifTrue: [ array at: 2 put: (self
				drawArrowOn: aCanvas
				at: vertices last
				from: self nextToLastPoint) ].

	(arrows == #back or: [arrows == #both])
		ifTrue: [ array at: 1 put: (self
				drawArrowOn: aCanvas
				at: vertices first
				from: self nextToFirstPoint) ].

	^array! !

!MUIPolygonMorph methodsFor: 'drawing' stamp: 'yo 9/2/2016 15:27'!
drawBorderOn: aCanvas 
	self
		drawClippedBorderOn: aCanvas
		usingEnds: (Array with: vertices first with: vertices last)! !

!MUIPolygonMorph methodsFor: 'drawing' stamp: 'yo 9/2/2016 15:27'!
drawBorderOn: aCanvas usingEnds: anArray 
	"Display my border on the canvas."
	"NOTE: Much of this code is also copied in  
	drawDashedBorderOn:  
	(should be factored)"
	| bigClipRect style |
	borderDashSpec
		ifNotNil: [^ self drawDashedBorderOn: aCanvas usingEnds: anArray].
	style := self borderStyle.
	bigClipRect := aCanvas clipRect expandBy: self borderWidth + 1 // 2.
	self
		lineSegmentsDo: [:p1 :p2 | | p2i p1i | 
			p1i := p1 asIntegerPoint.
			p2i := p2 asIntegerPoint.
			self hasArrows
				ifTrue: ["Shorten line ends so as not to interfere with tip  
					of arrow."
					((arrows == #back
								or: [arrows == #both])
							and: [p1 = vertices first])
						ifTrue: [p1i := anArray first asIntegerPoint].
					((arrows == #forward
								or: [arrows == #both])
							and: [p2 = vertices last])
						ifTrue: [p2i := anArray last asIntegerPoint]].
			(closed
					or: ["bigClipRect intersects: (p1i rect: p2i)  
						optimized:"
						((p1i min: p2i)
							max: bigClipRect origin)
							<= ((p1i max: p2i)
									min: bigClipRect corner)])
				ifTrue: [style
						drawLineFrom: p1i
						to: p2i
						on: aCanvas]]! !

!MUIPolygonMorph methodsFor: 'drawing' stamp: 'yo 9/2/2016 15:27'!
drawClippedBorderOn: aCanvas usingEnds: anArray 
	aCanvas clipBy: self bounds during:[:cc| self drawBorderOn: cc usingEnds: anArray].! !

!MUIPolygonMorph methodsFor: 'drawing' stamp: 'yo 9/2/2016 15:27'!
drawDashedBorderOn: aCanvas 
	self
		drawDashedBorderOn: aCanvas
		usingEnds: (Array with: vertices first with: vertices last)! !

!MUIPolygonMorph methodsFor: 'drawing' stamp: 'yo 9/2/2016 15:27'!
drawDashedBorderOn: aCanvas usingEnds: anArray 
	"Display my border on the canvas. NOTE: mostly copied from  
	drawBorderOn:"
	| bevel topLeftColor bottomRightColor bigClipRect lineColor segmentOffset |
	(borderColor isNil
			or: [borderColor isColor
					and: [borderColor isTransparent]])
		ifTrue: [^ self].
	lineColor := borderColor.
	bevel := false.
	"Border colors for bevelled effects depend on CW ordering of  
	vertices"
	borderColor == #raised
		ifTrue: [topLeftColor := color lighter.
			bottomRightColor := color darker.
			bevel := true].
	borderColor == #inset
		ifTrue: [topLeftColor := owner colorForInsets darker.
			bottomRightColor := owner colorForInsets lighter.
			bevel := true].
	bigClipRect := aCanvas clipRect expandBy: self borderWidth + 1 // 2.
	segmentOffset := self borderDashOffset.
	self
		lineSegmentsDo: [:p1 :p2 | | p1i p2i | 
			p1i := p1 asIntegerPoint.
			p2i := p2 asIntegerPoint.
			self hasArrows
				ifTrue: ["Shorten line ends so as not to interfere with tip  
					of arrow."
					((arrows == #back
								or: [arrows == #both])
							and: [p1 = vertices first])
						ifTrue: [p1i := anArray first asIntegerPoint].
					((arrows == #forward
								or: [arrows == #both])
							and: [p2 = vertices last])
						ifTrue: [p2i := anArray last asIntegerPoint]].
			(closed
					or: ["bigClipRect intersects: (p1i rect: p2i)  
						optimized:"
						((p1i min: p2i)
							max: bigClipRect origin)
							<= ((p1i max: p2i)
									min: bigClipRect corner)])
				ifTrue: [bevel
						ifTrue: [lineColor := (p1i quadrantOf: p2i)
											> 2
										ifTrue: [topLeftColor]
										ifFalse: [bottomRightColor]].
					segmentOffset := aCanvas
								line: p1i
								to: p2i
								width: borderWidth
								color: lineColor
								dashLength: borderDashSpec first
								secondColor: borderDashSpec third
								secondDashLength: borderDashSpec second
								startingOffset: segmentOffset]]! !

!MUIPolygonMorph methodsFor: 'drawing' stamp: 'yo 9/2/2016 15:27'!
drawOn: aCanvas 
	"Display the receiver, a spline curve, approximated by straight 
	line segments."
	| array |
	vertices size < 1
		ifTrue: [self error: 'a polygon must have at least one point'].
	closed ifTrue:
		[aCanvas drawPolygon: self getVertices fillStyle: self fillStyle.
		aCanvas isShadowDrawing ifTrue: [^ self]].
	array := self drawArrowsOn: aCanvas.
	self drawClippedBorderOn: aCanvas usingEnds: array.
! !

!MUIPolygonMorph methodsFor: 'drawing' stamp: 'yo 9/2/2016 15:27'!
drawOnFormCanvas: aCanvas 
	"Display the receiver, a spline curve, approximated by straight line segments."

	| |
	vertices size < 1 ifTrue: [self error: 'a polygon must have at least one point'].
	closed & color isTransparent not
		ifTrue: [aCanvas stencil: self filledForm at: bounds topLeft - 1 color: color].
	(borderColor isColor and: [borderColor isTranslucentColor])
		ifTrue: [aCanvas stencil: self borderForm at: bounds topLeft
						color: borderColor]
		ifFalse: [self drawBorderOn: aCanvas].
	self arrowForms do:
		[:f | aCanvas stencil: f at: f offset
			color: (borderColor isColor ifTrue: [borderColor] ifFalse: [color])]! !

!MUIPolygonMorph methodsFor: 'attachments' stamp: 'yo 9/2/2016 15:27'!
endShapeColor: aColor
	self borderColor: aColor.
	self isClosed ifTrue: [ self color: aColor ].! !

!MUIPolygonMorph methodsFor: 'attachments' stamp: 'yo 9/2/2016 15:27'!
endShapeWidth: aWidth
	| originalWidth originalVertices transform |
	originalWidth := self valueOfProperty: #originalWidth ifAbsentPut: [ self borderWidth isZero ifFalse: [ self borderWidth ] ifTrue: [ 2 ] ].
	self borderWidth: aWidth.
	originalVertices := self valueOfProperty: #originalVertices ifAbsentPut: [
		self vertices collect: [ :ea | (ea - (self referencePosition)) rotateBy: self heading degreesToRadians about: 0@0 ]
	].
	transform := MorphicTransform offset: 0@0 angle: self heading degreesToRadians scale: originalWidth / aWidth.
	self setVertices: (originalVertices collect: [ :ea |
		((transform transform: ea) + self referencePosition) asIntegerPoint
	]).
	self computeBounds.! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
extent: newExtent 
	"Not really advisable, but we can preserve most of the geometry if we don't
	shrink things too small."
	| safeExtent center |
	center := self referencePosition.
	safeExtent := newExtent max: 20@20.
	self setVertices: (vertices collect:
		[:p | p - center * (safeExtent asFloatPoint / (bounds extent max: 1@1)) + center])! !

!MUIPolygonMorph methodsFor: 'visual properties' stamp: 'yo 9/2/2016 15:27'!
fillStyle

	self isOpen
		ifTrue: [^ self borderColor  "easy access to line color from halo"]
		ifFalse: [^ super fillStyle]! !

!MUIPolygonMorph methodsFor: 'visual properties' stamp: 'yo 9/2/2016 15:27'!
fillStyle: newColor

	self isOpen
		ifTrue: [^ self borderColor: newColor asColor "easy access to line color from halo"]
		ifFalse: [^ super fillStyle: newColor]! !

!MUIPolygonMorph methodsFor: 'private' stamp: 'yo 9/2/2016 15:27'!
filledForm
	"Note: The filled form is actually 2 pixels bigger than bounds, and the point corresponding to this morphs' position is at 1@1 in the form.  This is due to the details of the fillig routines, at least one of which requires an extra 1-pixel margin around the outside.  Computation of the filled form is done only on demand."
	| bb origin |
	closed ifFalse: [^ filledForm := nil].
	filledForm ifNotNil: [^ filledForm].
	filledForm := Form extent: bounds extent+2.

	"Draw the border..."
	bb := (BitBlt toForm: filledForm) sourceForm: nil; fillColor: Color black;
			combinationRule: Form over; width: 1; height: 1.
	origin := bounds topLeft asIntegerPoint-1.
	self lineSegmentsDo: [:p1 :p2 | bb drawFrom: p1 asIntegerPoint-origin
										to: p2 asIntegerPoint-origin].

	"Fill it in..."
	filledForm convexShapeFill: Color black.

	(borderColor isColor and: [borderColor isTranslucentColor]) ifTrue:
		["If border is stored as a form, then erase any overlap now."
		filledForm copy: self borderForm boundingBox from: self borderForm
			to: 1@1 rule: Form erase].

	^ filledForm! !

!MUIPolygonMorph methodsFor: 'attachments' stamp: 'yo 9/2/2016 15:27'!
firstVertex
	^vertices first! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
flipHAroundX: centerX
	"Flip me horizontally around the center.  If centerX is nil, compute my center of gravity."

	| cent |
	cent := centerX 
		ifNil: [bounds center x
			"cent := 0.
			vertices do: [:each | cent := cent + each x].
			cent asFloat / vertices size"]		"average is the center"
		ifNotNil: [centerX].
	self setVertices: (vertices collect: [:vv |
			((vv x - cent) * -1 + cent) @ vv y]) reversed.! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
flipVAroundY: centerY
	"Flip me vertically around the center.  If centerY is nil, compute my center of gravity."

	| cent |
	cent := centerY 
		ifNil: [bounds center y
			"cent := 0.
			vertices do: [:each | cent := cent + each y].
			cent asFloat / vertices size"]		"average is the center"
		ifNotNil: [centerY].
	self setVertices: (vertices collect: [:vv |
			vv x @ ((vv y - cent) * -1 + cent)]) reversed.! !

!MUIPolygonMorph methodsFor: 'private' stamp: 'yo 9/2/2016 15:27'!
getVertices

	smoothCurve ifFalse: [^ vertices].

	"For curves, enumerate the full set of interpolated points"
	^ Array streamContents:
		[:s | self lineSegmentsDo: [:pt1 :pt2 | s nextPut: pt1]]! !

!MUIPolygonMorph methodsFor: 'editing' stamp: 'yo 9/2/2016 15:27'!
handleColorAt: vertIndex
      "This is a backstop for MixedCurveMorph"

^ Color yellow
! !

!MUIPolygonMorph methodsFor: 'event handling' stamp: 'yo 9/2/2016 16:27'!
handlesMouseDown: evt
	^ false
! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
handlesShowingPhrase
	^ (self showingHandles
		ifTrue: ['hide handles']
		ifFalse: ['show handles']) translated! !

!MUIPolygonMorph methodsFor: 'testing' stamp: 'yo 9/2/2016 15:27'!
hasArrows
	"Are all the conditions meet for having arrows?"
	^ (closed
		or: [arrows == #none
				or: [vertices size < 2]]) not! !

!MUIPolygonMorph methodsFor: 'private' stamp: 'yo 9/2/2016 15:27'!
includesHandle: aMorph

	handles ifNil: [^ false].
	^ handles includes: aMorph! !

!MUIPolygonMorph methodsFor: 'initialization' stamp: 'yo 9/2/2016 15:27'!
initialize
"initialize the state of the receiver"
	super initialize.
""
	vertices := Array
				with: 5 @ 0
				with: 20 @ 10
				with: 0 @ 20.
	closed := true.
	smoothCurve := false.
	arrows := #none.
	self computeBounds! !

!MUIPolygonMorph methodsFor: 'editing' stamp: 'yo 9/2/2016 15:27'!
insertVertexAt: anIndex put: aValue
	"This serves as a hook and a backstop for MixedCurveMorph."
	self setVertices: (vertices copyReplaceFrom: anIndex + 1 to: anIndex 
									with: (Array with: aValue)).! !

!MUIPolygonMorph methodsFor: 'debug and other' stamp: 'yo 9/2/2016 15:27'!
installModelIn: aWorld
	aWorld isWorldMorph ifTrue: [self addHandles]! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
intersectionWithLineSegmentFromCenterTo: aPoint 
	^self closestPointTo: aPoint! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
intersectionsWith: aRectangle
	"Answer a Set of points where the given Rectangle intersects with me.
	Ignores arrowForms."

	| retval |
	retval := IdentitySet new: 4.
	(self bounds intersects: aRectangle) ifFalse: [^ retval].

	self lineSegmentsDo: [ :lp1 :lp2 | | polySeg |
		polySeg := LineSegment from: lp1 to: lp2.
		aRectangle lineSegmentsDo: [ :rp1 :rp2 | | rectSeg int |
			rectSeg := LineSegment from: rp1 to: rp2.
			int := polySeg intersectionWith: rectSeg.
			int ifNotNil: [ retval add: int ].
		].
	].

	^retval
! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
intersects: aRectangle 
	"Answer whether any of my segments intersects aRectangle, which is in World coordinates."
	| rect |
	(super intersects: aRectangle) ifFalse: [ ^false ].
	rect := self bounds: aRectangle in: self world.
	self
		lineSegmentsDo: [:p1 :p2 | (rect intersectsLineFrom: p1 to: p2)
				ifTrue: [^ true]].
	^ false! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
isBordered
	^false! !

!MUIPolygonMorph methodsFor: 'access' stamp: 'yo 9/2/2016 15:27'!
isClosed
	^ closed! !

!MUIPolygonMorph methodsFor: 'access' stamp: 'yo 9/2/2016 15:27'!
isCurve
	^ smoothCurve! !

!MUIPolygonMorph methodsFor: 'testing' stamp: 'yo 9/2/2016 15:27'!
isCurvier
	"Test used by smoothing routines.  If true use true closed curve splines for closed curves. If not mimic old stodgy curveMorph curves with one sharp bend.. Override this routine in classes where backward compatability is still needed."
	^ CurvierMorph drawCurvier! !

!MUIPolygonMorph methodsFor: 'testing' stamp: 'yo 9/2/2016 15:27'!
isCurvy
	"Test for significant curves.  
	Small smoothcurves in practice are straight."
	^ smoothCurve
		and: [vertices size > 2]! !

!MUIPolygonMorph methodsFor: 'testing' stamp: 'yo 9/2/2016 15:27'!
isLineMorph
	^closed not! !

!MUIPolygonMorph methodsFor: 'access' stamp: 'yo 9/2/2016 15:27'!
isOpen
	^ closed not! !

!MUIPolygonMorph methodsFor: 'dropping/grabbing' stamp: 'yo 9/2/2016 15:27'!
justDroppedInto: newOwner event: evt

	| delta |
	(newOwner isKindOf: PasteUpMorph) ifTrue:
		["Compensate for border width so that gridded drop
			is consistent with gridded drag of handles."
		delta := borderWidth+1//2.
		self position: (newOwner gridPoint: self position + delta) - delta].
	^ super justDroppedInto: newOwner event: evt! !

!MUIPolygonMorph methodsFor: 'attachments' stamp: 'yo 9/2/2016 15:27'!
lastVertex
	^vertices last! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
lineBorderColor
	^self borderColor! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
lineBorderColor: aColor
	self borderColor: aColor! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
lineBorderWidth

	^self borderWidth! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
lineBorderWidth: anInteger

	self borderWidth: anInteger! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
lineColor
	^self borderColor! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
lineColor: aColor
	self borderColor: aColor! !

!MUIPolygonMorph methodsFor: 'private' stamp: 'yo 9/2/2016 15:27'!
lineSegments
	| lineSegments |
	lineSegments := OrderedCollection new.
	self lineSegmentsDo: [:p1 :p2 | lineSegments addLast: (Array with: p1 with: p2)].
	^ lineSegments! !

!MUIPolygonMorph methodsFor: 'smoothing' stamp: 'yo 9/2/2016 15:27'!
lineSegmentsDo: endPointsBlock 
	"Emit a sequence of segment endpoints into endPointsBlock."
	"Unlike the method this one replaces we expect the curve 
	coefficents not the dirivatives"
	"Also unlike the replaced method the smooth closed curve
	does 
	not need an extra vertex. 
	We take care of the extra endpoint here. Just like for 
	segmented curves."
	| cs x y beginPoint |
	vertices size < 1
		ifTrue: [^ self].
	"test too few vertices first"
	self isCurvy
		ifFalse: [beginPoint := nil.
			"smoothCurve 
			ifTrue: [cs := self coefficients]."
			"some things still depend on smoothCurves having 
			curveState"
			vertices
				do: [:vert | 
					beginPoint
						ifNotNil: [endPointsBlock value: beginPoint value: vert].
					beginPoint := vert].
			(closed
					or: [vertices size = 1])
				ifTrue: [endPointsBlock value: beginPoint value: vertices first].
			^ self].
	"For curves we include all the interpolated sub segments."
	"self assert: [(vertices size > 2 )].	"
	cs := self coefficients.
	beginPoint := (x := cs first first) @ (y := cs fifth first).
	(closed
		ifTrue: [1 to: cs first size]
		ifFalse: [1 to: cs first size - 1])
		do: [:i | | x1 y1 endPoint n y2 t x3 y3 x2 | 
			"taylor series coefficients"
			x1 := cs second at: i.
			y1 := cs sixth at: i.
			x2 := cs third at: i.
			y2 := cs seventh at: i.
			x3 := cs fourth at: i.
			y3 := cs eighth at: i.
			n := cs ninth at: i.
			"guess n 
			n := 5 max: (x2 abs + y2 abs * 2.0 + (cs third atWrap:
			i 
			+ 1) abs + (cs seventh atWrap: i + 1) abs / 100.0) 
			rounded."
			1
				to: n - 1
				do: [:j | 
					t := j asFloat / n asFloat.
					endPoint := x3 * t + x2 * t + x1 * t + x @ (y3 * t + y2 * t + y1 * t + y).
					endPointsBlock value: beginPoint value: endPoint.
					beginPoint := endPoint].
			endPoint := (x := cs first atWrap: i + 1) @ (y := cs fifth atWrap: i + 1).
			endPointsBlock value: beginPoint value: endPoint.
			beginPoint := endPoint]! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
lineWidth

	^self borderWidth! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
lineWidth: anInteger

	self borderWidth: (anInteger rounded max: 1)! !

!MUIPolygonMorph methodsFor: 'caching' stamp: 'yo 9/2/2016 15:27'!
loadCachedState
	"Prepare for fast response -- next page of a book?"
	self filledForm.
	self arrowForms! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
makeBackArrow
	arrows := #back.
	self computeBounds! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
makeBothArrows
	arrows := #both.
	self computeBounds! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
makeClosed
	closed := true.
	handles ifNotNil: [self removeHandles; addHandles].
	self computeBounds! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
makeForwardArrow
	arrows := #forward.
	self computeBounds! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
makeNoArrows
	arrows := #none.
	self computeBounds! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
makeOpen
	closed := false.
	handles ifNotNil: [self removeHandles; addHandles].
	self computeBounds! !

!MUIPolygonMorph methodsFor: 'access' stamp: 'yo 9/2/2016 15:27'!
makeOpenOrClosed
	"toggle the open/closed status of the receiver"
	closed ifTrue: [self makeOpen] ifFalse: [self makeClosed]! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
merge: aPolygon 
	"Expand myself to enclose the other polygon.  (Later merge overlapping or disjoint in a smart way.)  For now, the two polygons must share at least two vertices.  Shared vertices must come one after the other in each polygon.  Polygons must not overlap."

	| shared mv vv hv xx |
	shared := vertices select: [:mine | aPolygon vertices includes: mine].
	shared size < 2 ifTrue: [^nil].	"not sharing a segment"
	mv := vertices asOrderedCollection.
	[shared includes: mv first] whileFalse: 
			["rotate them"

			vv := mv removeFirst.
			mv addLast: vv].
	hv := aPolygon vertices asOrderedCollection.
	[mv first = hv first] whileFalse: 
			["rotate him until same shared vertex is first"

			vv := hv removeFirst.
			hv addLast: vv].
	[shared size > 2] whileTrue: 
			[shared := shared asOrderedCollection.
			(self 
				mergeDropThird: mv
				in: hv
				from: shared) ifNil: [^nil]].
	"works by side effect on the lists"
	(mv second) = hv last 
		ifTrue: 
			[mv
				removeFirst;
				removeFirst.
			^self setVertices: (hv , mv) asArray].
	(hv second) = mv last 
		ifTrue: 
			[hv
				removeFirst;
				removeFirst.
			^self setVertices: (mv , hv) asArray].
	(mv second) = (hv second) 
		ifTrue: 
			[hv removeFirst.
			mv remove: (mv second).
			xx := mv removeFirst.
			^self setVertices: (hv , (Array with: xx) , mv reversed) asArray].
	mv last = hv last 
		ifTrue: 
			[mv removeLast.
			hv removeFirst.
			^self setVertices: (mv , hv reversed) asArray].
	^nil! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
mergeDropThird: mv in: hv from: shared 
	"We are merging two polygons.  In this case, they have at least three identical shared vertices.  Make sure they are sequential in each, and drop the middle one from vertex lists mv, hv, and shared.  First vertices on lists are identical already."

	"know (mv first = hv first)"

	| mdrop vv |
	(shared includes: (mv at: mv size - 2)) 
		ifTrue: [(shared includes: mv last) ifTrue: [mdrop := mv last]]
		ifFalse: 
			[(shared includes: mv last) 
				ifTrue: [(shared includes: mv second) ifTrue: [mdrop := mv first]]].
	(shared includes: (mv third)) 
		ifTrue: [(shared includes: mv second) ifTrue: [mdrop := mv second]].
	mdrop ifNil: [^nil].
	mv remove: mdrop.
	hv remove: mdrop.
	shared remove: mdrop.
	[shared includes: mv first] whileFalse: 
			["rotate them"

			vv := mv removeFirst.
			mv addLast: vv].
	[mv first = hv first] whileFalse: 
			["rotate him until same shared vertex is first"

			vv := hv removeFirst.
			hv addLast: vv]! !

!MUIPolygonMorph methodsFor: 'access' stamp: 'yo 9/2/2016 15:27'!
midVertices
	"Return and array of midpoints for this line or closed curve"
	| midPts nextVertIx tweens |
	vertices size < 2
		ifTrue: [^ vertices].
	midPts := OrderedCollection new.
	nextVertIx := 2.
	tweens := OrderedCollection new.
	tweens add: vertices first asIntegerPoint.
	"guarantee at least two points."
	self
		lineSegmentsDo: [:p1 :p2 | 
			tweens addLast: p2 asIntegerPoint.
			p2
					= (vertices atWrap: nextVertIx)
				ifTrue: ["Found endPoint."
					midPts addLast: (tweens atWrap: tweens size + 1 // 2)
							+ (tweens at: tweens size // 2 + 1) // 2.
					"wiz 6/19/2004 20:11 adjusted to handle  
					one segment properly"
					tweens := OrderedCollection new.
					tweens add: p2 asIntegerPoint.
					"guarantee at least two points."
					nextVertIx := nextVertIx + 1]].
	^ midPts asArray! !

!MUIPolygonMorph methodsFor: 'attachments' stamp: 'yo 9/2/2016 15:27'!
midpoint
	"Answer the midpoint along my segments"
	| middle |
	middle := self totalLength.
	middle < 2 ifTrue: [ ^ self center ].
	middle := middle / 2.
	self lineSegmentsDo: [ :a :b | | dist |
		dist := (a dist: b).
		middle < dist
			ifTrue: [ ^(a + ((b - a) * (middle / dist))) asIntegerPoint ].
		middle := middle - dist.
	].
	self error: 'can''t happen'! !

!MUIPolygonMorph methodsFor: 'editing' stamp: 'yo 9/2/2016 15:27'!
newVertex: ix event: evt fromHandle: handle
	"Insert a new vertex and fix everything up!! Install the drag-handle of the new vertex as recipient of further mouse events."

	| pt |
	"(self hasProperty: #noNewVertices) ifFalse:
		[pt := evt cursorPoint.
		self setVertices: (vertices copyReplaceFrom: ix + 1 to: ix with: (Array with: pt)).
		evt hand newMouseFocus: (handles at: ((ix + 1) * 2) - 1)]"
	"modified to remove now vestigial test. see PolygonMorph class>>arrowprototype"
	pt := evt cursorPoint.
	self  insertVertexAt: ix put:  pt .
	evt hand newMouseFocus: (handles at: ((ix + 1) * 2) - 1)! !

!MUIPolygonMorph methodsFor: 'editing' stamp: 'yo 9/2/2016 15:27'!
newVertex: arg1 fromHandle: arg2 afterVert: arg3
	"Reorder the arguments for existing event handlers"
	(arg3 isMorph and:[arg3 eventHandler notNil]) ifTrue:[arg3 eventHandler fixReversedValueMessages].
	^self newVertex: arg1 event: arg2 fromHandle: arg3! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
nextDuplicateVertexIndex
	vertices
		doWithIndex: [:vert :index | ((index between: 2 and: vertices size - 1)
					and: [| epsilon v1 v2 | 
						v1 := vertices at: index - 1.
						v2 := vertices at: index + 1.
						epsilon := ((v1 x - v2 x) abs max: (v1 y - v2 y) abs)
									// 32 max: 1.
						vert
							onLineFrom: v1
							to: v2
							within: epsilon])
				ifTrue: [^ index]].
	^ 0! !

!MUIPolygonMorph methodsFor: 'smoothing' stamp: 'yo 9/2/2016 15:27'!
nextToFirstPoint
	"For arrow direction"
	self isCurvy
		ifTrue: [curveState
				ifNil: [self coefficients].
			^ curveState second]
		ifFalse: [^ vertices second]! !

!MUIPolygonMorph methodsFor: 'smoothing' stamp: 'yo 9/2/2016 15:27'!
nextToLastPoint
	"For arrow direction"
	self isCurvy
		ifTrue: [curveState
				ifNil: [self coefficients].
			^ curveState third]
		ifFalse: [^ vertices at: vertices size - 1]! !

!MUIPolygonMorph methodsFor: 'access' stamp: 'yo 9/2/2016 15:27'!
openOrClosePhrase
	| curveName |
	curveName := (self isCurve
				ifTrue: ['curve']
				ifFalse: ['polygon']) translated.
	^ closed
		ifTrue: ['make open {1}' translated format: {curveName}]
		ifFalse: ['make closed {1}' translated format: {curveName}]! !

!MUIPolygonMorph methodsFor: 'private' stamp: 'yo 9/2/2016 15:27'!
privateMoveBy: delta
	super privateMoveBy: delta.
	vertices := vertices collect: [:p | p + delta].
	self arrowForms do: [:f | f offset: f offset + delta].
	curveState := nil.  "Force recomputation"
	(self valueOfProperty: #referencePosition) ifNotNil:
		[:oldPos | self setProperty: #referencePosition toValue: oldPos + delta]! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
quickFill: ignored! !

!MUIPolygonMorph methodsFor: 'shaping' stamp: 'yo 9/2/2016 15:27'!
rectOval
	"Set my vertices to an array of corner vertices.
	Order of vertices is in the tradion of warpblt quads."

	self setVertices: self bounds corners.! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
reduceVertices
	"Reduces the vertices size, when 3 vertices are on the same line with a 
	little epsilon. Based on code by Steffen Mueller"
	| dup |
	[ (dup := self nextDuplicateVertexIndex) > 0 ] whileTrue: [
		self setVertices: (vertices copyWithoutIndex: dup)
	].
	^vertices size.! !

!MUIPolygonMorph methodsFor: 'geometry eToy' stamp: 'yo 9/2/2016 15:27'!
referencePosition 
	"Return the current reference position of the receiver"
	^ self valueOfProperty: #referencePosition ifAbsent: [super referencePosition]
! !

!MUIPolygonMorph methodsFor: 'caching' stamp: 'yo 9/2/2016 15:27'!
releaseCachedState
	super releaseCachedState.
	filledForm := nil.
	arrowForms := nil.
	borderForm := nil.
	curveState := nil.
	(self hasProperty: #flex) ifTrue:
		[self removeProperty: #unflexedVertices;
			removeProperty: #flex].
! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
removeHandles
	"tk 9/2/97 allow it to be called twice (when nil already)"

	handles ifNotNil: [
		handles do: [:h | h delete].
		handles := nil].! !

!MUIPolygonMorph methodsFor: 'dashes' stamp: 'yo 9/2/2016 15:27'!
removeVertex: aVert
	"Make sure that I am not left with less than two vertices"
	| newVertices |
	vertices size < 2 ifTrue: [ ^self ].
	newVertices := vertices copyWithout: aVert.
	newVertices size caseOf: {
		[1] -> [ newVertices := { newVertices first . newVertices first } ].
		[0] -> [ newVertices := { aVert . aVert } ]
	} otherwise: [].
	self setVertices: newVertices 
! !

!MUIPolygonMorph methodsFor: 'debug and other' stamp: 'yo 9/2/2016 15:27'!
rotateTestFlip: aBool 
	"Return one copy of me for each vertex using each vertex as  
	the  
	starting point.  
	Vary to border color to destinguish the copies.  
	This tests closed curves for their consistency.  
	The flip boolean tests the reversed rotations."
	| len colors verts flip |
	verts := self vertices.
	flip := aBool == true
				ifTrue: [1]
				ifFalse: [0].
	len := verts size.
	colors := Color wheel: len*2 .
	(1 to: len)
		do: [:i | | j | (self copy
				borderColor: (colors at: (j:=i * 2 - flip));
				 yourself)
				setVertices: (verts flipRotated: j);
				 openInWorld]! !

!MUIPolygonMorph methodsFor: 'geometry eToy' stamp: 'yo 9/2/2016 15:27'!
rotationCenter
	"Return the rotation center of the receiver. The rotation center defines the relative offset inside the receiver's bounds for locating the reference position."
	| refPos |
	refPos := self valueOfProperty: #referencePosition
		ifAbsent: [^ 0.5@0.5].
	^ (refPos - self bounds origin) / self bounds extent asFloatPoint! !

!MUIPolygonMorph methodsFor: 'geometry eToy' stamp: 'yo 9/2/2016 15:27'!
rotationCenter: aPointOrNil
	"Set the new rotation center of the receiver. The rotation center defines the relative offset inside the receiver's bounds for locating the reference position."
	| box |
	aPointOrNil isNil
		ifTrue: [self removeProperty: #referencePosition]
		ifFalse: [box := self bounds.
				self setProperty: #referencePosition
					toValue: box origin + (aPointOrNil * box extent)]
! !

!MUIPolygonMorph methodsFor: 'rotate scale and flex' stamp: 'yo 9/2/2016 15:27'!
rotationDegrees

	^ self forwardDirection! !

!MUIPolygonMorph methodsFor: 'halo control' stamp: 'yo 9/2/2016 15:27'!
rotationDegrees: degrees 
	| flex center |
	(center := self valueOfProperty: #referencePosition) ifNil:
		[self setProperty: #referencePosition toValue: (center := self bounds center)].
	flex := (MorphicTransform offset: center negated)
			withAngle: (degrees - self forwardDirection) degreesToRadians.
	self setVertices: (vertices collect: [:v | (flex transform: v) - flex offset]).
	self forwardDirection: degrees.

! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
setRotationCenterFrom: aPoint
	"Polygons store their referencePosition."
	self setProperty: #referencePosition toValue: aPoint! !

!MUIPolygonMorph methodsFor: 'private' stamp: 'yo 9/2/2016 15:27'!
setVertices: newVertices
	vertices := newVertices.
	handles ifNotNil: [self removeHandles; addHandles].
	self computeBounds! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
showOrHideHandles
	self showingHandles
		ifTrue:	[self removeHandles]
		ifFalse:	[self addHandles]! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
showingHandles
	^ handles notNil! !

!MUIPolygonMorph methodsFor: 'smoothing' stamp: 'yo 9/2/2016 15:27'!
slopes: knots 
	"Choose slopes according to state of polygon and preferences"
	self isCurvy
		ifFalse: [^ knots segmentedSlopes].
	^ (closed
			and: [self isCurvier])
		ifTrue: [knots closedCubicSlopes]
		ifFalse: [knots naturalCubicSlopes]! !

!MUIPolygonMorph methodsFor: 'access' stamp: 'yo 9/2/2016 15:27'!
smoothOrSegmentedPhrase
				| lineName |
	lineName := (closed
						ifTrue: ['outline']
						ifFalse: ['line']) translated.

			^ self isCurve
				ifTrue: ['make segmented {1}' translated format: {lineName}]
				ifFalse: ['make smooth {1}' translated format: {lineName}].! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
specifyDashedLine

	| executableSpec newSpec |
	executableSpec := UIManager default
		request:
'Enter a dash specification as
{ major dash length. minor dash length. minor dash color }
The major dash will have the normal border color.
A blank response will remove the dash specification.
[Note: You may give 5 items as, eg, {10. 5. Color white. 0. 3}
where the 4th ityem is zero, and the 5th is the number of pixels
by which the dashes will move in each step of animation]' translated
		initialAnswer: '{ 10. 5. Color red }'.
	executableSpec isEmpty ifTrue:
		[^ self stopStepping; dashedBorder: nil].
	newSpec := [Compiler evaluate: executableSpec] ifError:
		[^ self stopStepping; dashedBorder: nil].
	newSpec first isNumber & newSpec second isNumber & newSpec third isColor ifFalse:
		[^ self stopStepping; dashedBorder: nil].
	newSpec size = 3 ifTrue:
		[^ self stopStepping; dashedBorder: newSpec].
	(newSpec size = 5 and: [newSpec fourth isNumber & newSpec fifth isNumber]) ifTrue:
		[^ self dashedBorder: newSpec; startStepping].
! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
standardArrows

	self removeProperty: #arrowSpec.
	self computeBounds! !

!MUIPolygonMorph methodsFor: 'stepping and presenter' stamp: 'yo 9/2/2016 15:27'!
step
	borderDashSpec ifNil: [^super step].
	borderDashSpec size < 5 ifTrue: [^super step].

	"Only for dashed lines with creep"
	borderDashSpec at: 4 put: (borderDashSpec fourth) + borderDashSpec fifth.
	self changed.
	^super step! !

!MUIPolygonMorph methodsFor: 'testing' stamp: 'yo 9/2/2016 15:27'!
stepTime

	^ 100! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
straighten
	self setVertices: { vertices first . vertices last }! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
toggleHandles

	handles ifNil: [self addHandles] ifNotNil: [self removeHandles].

! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
toggleSmoothing

	smoothCurve := smoothCurve not.
	handles ifNotNil: [self removeHandles; addHandles].
	self computeBounds! !

!MUIPolygonMorph methodsFor: 'attachments' stamp: 'yo 9/2/2016 15:27'!
totalLength
	"Answer the full length of my segments. Can take a long time if I'm curved."
	| length |
	length := 0.
	self lineSegmentsDo: [ :a :b | length := length + (a dist: b) ].
	^length.! !

!MUIPolygonMorph methodsFor: 'private' stamp: 'yo 9/2/2016 15:27'!
transformVerticesFrom: oldOwner to: newOwner
	| oldTransform newTransform world newVertices |
	world := self world.
	oldTransform := oldOwner
		ifNil: [ IdentityTransform new ]
		ifNotNil: [ oldOwner transformFrom: world ].
	newTransform := newOwner
		ifNil: [ IdentityTransform new ]
		ifNotNil: [ newOwner transformFrom: world ].
	newVertices := vertices collect: [ :ea | newTransform globalPointToLocal:
		(oldTransform localPointToGlobal: ea) ].
	self setVertices: newVertices.
! !

!MUIPolygonMorph methodsFor: 'geometry' stamp: 'yo 9/2/2016 15:27'!
transformedBy: aTransform
	self setVertices: (self vertices collect:[:v| aTransform localPointToGlobal: v])! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
unrotatedLength
	"If the receiver bears rotation without a transformation morph, answer what its length in the direction it is headed is"

	vertices size = 2 ifTrue:
		[^ (vertices second - vertices first) r].

	^ ((PolygonMorph new setVertices: vertices) rotationDegrees: self rotationDegrees negated) height! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
unrotatedLength: aLength
	"If the receiver bears rotation without a transformation morph, answer what its length in the direction it is headed is"

	vertices size = 2 ifTrue: [^ self arrowLength: aLength].

	self setVertices: ((((PolygonMorph new setVertices: vertices) rotationDegrees: self rotationDegrees negated) height: aLength) rotationDegrees: 0) vertices! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
unrotatedWidth
	"If the receiver bears rotation without a transformation morph, answer what its length in the direction it is headed is"
	
	vertices size = 2 ifTrue: [^ self borderWidth].
	^ ((PolygonMorph new setVertices: vertices) rotationDegrees: self rotationDegrees negated) width! !

!MUIPolygonMorph methodsFor: 'menu' stamp: 'yo 9/2/2016 15:27'!
unrotatedWidth: aWidth
	"If the receiver bears rotation without a transformation morph, answer what its length in the direction it is headed is"

	self borderWidth: aWidth! !

!MUIPolygonMorph methodsFor: 'editing' stamp: 'yo 9/2/2016 15:27'!
updateHandles
	self isCurvy
		ifTrue: [handles first center: vertices first.
			handles last center: vertices last.
			self midVertices
				withIndexDo: [:midPt :vertIndex | (closed
							or: [vertIndex < vertices size])
						ifTrue: [| newVert |
							newVert := handles atWrap: vertIndex * 2.
							newVert position: midPt - (newVert extent // 2)]]]
		ifFalse: [vertices
				withIndexDo: [:vertPt :vertIndex |
					| oldVert | 
					oldVert := handles at: vertIndex * 2 - 1.
					oldVert position: vertPt - (oldVert extent // 2).
					(closed
							or: [vertIndex < vertices size])
						ifTrue: [| newVert |
							newVert := handles at: vertIndex * 2.
							newVert position: vertPt
									+ (vertices atWrap: vertIndex + 1) - newVert extent // 2 + (1 @ -1)]]]! !

!MUIPolygonMorph methodsFor: 'dashes' stamp: 'yo 9/2/2016 15:27'!
vertexAt: n
	^vertices at: (n min: vertices size).! !

!MUIPolygonMorph methodsFor: 'access' stamp: 'yo 9/2/2016 15:27'!
vertices
	^ vertices! !

!MUIPolygonMorph methodsFor: 'initialization' stamp: 'yo 9/2/2016 15:27'!
vertices: verts color: aColor borderWidth: borderWidthInteger borderColor: anotherColor 
	super initialize.
""
	vertices := verts.
	color := aColor.
	borderWidth := borderWidthInteger.
	borderColor := anotherColor.
	closed := vertices size > 2.
	arrows := #none.
	self computeBounds! !

!MUIPolygonMorph methodsFor: 'editing' stamp: 'yo 9/2/2016 15:27'!
verticesAt: ix put: newPoint
	vertices at: ix put: newPoint.
	self computeBounds! !

!MUIPolygonMorph methodsFor: 'testing' stamp: 'yo 9/2/2016 15:27'!
wantsSteps
	super wantsSteps ifTrue: [^true].

	"For crawling ants effect of dashed line."
	borderDashSpec ifNil: [^false].
	^borderDashSpec size = 5 and: [(borderDashSpec fifth) > 0]! !

Morph subclass: #MUIEnvelopeEditorMorph
	instanceVariableNames: 'isSelecting selectionStart selection selectionPos curve editingCurve samples newSamples selectionRect selectedSamples selectionInterval selectedControlPoints windowLeft windowRight'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MUI'!

!MUIEnvelopeEditorMorph class methodsFor: 'as yet unclassified' stamp: 'yo 9/6/2016 16:32'!
fromHandFreehand
	"Let the user draw a polygon, holding the mouse down, and ending
		by clicking within 5 of the first point..."
	"
	MUIEnvelopeEditorMorph fromHandFreehand
	"
	| p1 poly pN opposite editor min |
	Cursor crossHair showWhile:
		[[Sensor anyButtonPressed] whileFalse:
			[self currentWorld displayWorldSafely; runStepMethods].
		p1 := Sensor cursorPoint].
	opposite := (Display colorAt: p1) negated.
	opposite = Color transparent ifTrue: [opposite := Color red].
	(poly := MUIPolygonMorph vertices: {p1} color: opposite borderWidth: 2 borderColor: Color gray) openInWorld.
	self currentWorld displayWorldSafely; runStepMethods.
	[Sensor anyButtonPressed] whileTrue:
			[pN := Sensor cursorPoint.
			(pN x > (poly vertices last x + 10)) ifTrue:
				[poly setVertices: (poly vertices copyWith: pN).
				self currentWorld displayWorldSafely; runStepMethods]].
	editor := MUIEnvelopeEditorMorph new.
	editor curve: poly.
	min := poly vertices min.
	poly beSmoothCurve.
	poly setVertices: (poly vertices - min).
	editor openInWorld.
	^ editor
! !

!MUIEnvelopeEditorMorph methodsFor: 'curve based solver' stamp: 'yo 8/29/2016 16:13'!
adjustWindow: window with: diffY

	| width x windowFirst curve ds proposedDs newHeight origHeight ratio newProposedS now newProposedRemaining remaining base |
	diffY = 0 ifTrue: [^ #()].
	windowFirst := window first.

	base := (samples at: window first - 1) y.
	width := window size asFloat.
	ds := (0 to: window size - 1) collect: [:w | (samples at: w + windowFirst) y - (samples at: (w - 1) + windowFirst) y].

	"We take the suggested curve, which is just linearly scaled version of the curve in the window and for each differential, we multiply that with a non-linear curve to 'allow' the part that is far away from the boundary to change more freely."

	curve := (1 to: window size) collect: [:w | 
		x := w <= (window size / 2)
			ifTrue: [w * 2 / width]
			ifFalse: [window size - w / (width / 2)].
		x * x * x].

	origHeight := ((samples at: window last) - (samples at: window first - 1)) y.
	newHeight := origHeight + diffY.
	origHeight ~= 0 ifTrue: [
		proposedDs := ds * (newHeight / origHeight asFloat)]
	ifFalse: [
		proposedDs := (1 to: window size) collect: [:w | newHeight / width]].

	remaining := proposedDs - ds.

	newProposedRemaining := remaining * curve.
	ratio := remaining sum / newProposedRemaining sum.

	newProposedRemaining := newProposedRemaining * ratio.

	now := 0.
	newProposedS := (1 to: window size) collect: [:w |
		now := now + ((ds at: w) + (newProposedRemaining at: w))].

	^ window collect: [:s |
		(samples at: s) x@ (((newProposedS at: (s - windowFirst + 1)) + base))].
! !

!MUIEnvelopeEditorMorph methodsFor: 'curve based solver' stamp: 'yo 8/30/2016 10:01'!
adjustWindowLeft: diffY

	| window windowFirst |
	windowFirst := 1 max: (selectedSamples first first - windowSize).
	window := windowFirst to: selectedSamples first first - 1.
	^ self adjustWindow: window with: diffY.
	! !

!MUIEnvelopeEditorMorph methodsFor: 'curve based solver' stamp: 'yo 8/30/2016 10:01'!
adjustWindowRight: diffY

	| window windowFirst |
	windowFirst := 1 max: (selectedSamples last first + 1).
	window := windowFirst to: windowFirst + windowSize - 1.
	^ self adjustWindowRight: window with: diffY
! !

!MUIEnvelopeEditorMorph methodsFor: 'curve based solver' stamp: 'yo 8/30/2016 10:01'!
adjustWindowRight: window with: diffY

	| width x windowFirst curve ds proposedDs newHeight origHeight ratio newProposedS now newProposedRemaining remaining base |
	diffY = 0 ifTrue: [^ #()].
	windowFirst := window first.

	base := (newSamples at: window first - 1) y.
	width := window size asFloat.
	ds := (1 to: window size) collect: [:w | (samples at: w + windowFirst) y - (samples at: (w - 1) + windowFirst) y].

	"We take the suggested curve, which is just linearly scaled version of the curve in the window and for each differential, we multiply that with a non-linear curve to 'allow' the part that is far away from the boundary to change more freely."

	curve := (1 to: window size) collect: [:w | 
		x := w <= (window size / 2)
			ifTrue: [w * 2 / width]
			ifFalse: [window size - w / (width / 2)].
		x * x * x].

	origHeight := ((samples at: window first) - (samples at: window last + 1)) y.
	newHeight := origHeight + diffY.
	
	origHeight ~= 0 ifTrue: [
		proposedDs := ds * (newHeight / origHeight asFloat)]
	ifFalse: [
		proposedDs := (1 to: window size) collect: [:w | newHeight / width negated]].

	remaining := proposedDs - ds.

	newProposedRemaining := remaining * curve.
	ratio := remaining sum / newProposedRemaining sum.

	newProposedRemaining := newProposedRemaining * ratio.

	now := 0.
	newProposedS := (1 to: window size) collect: [:w |
		now := now + ((ds at: w) + (newProposedRemaining at: w))].

	^ window collect: [:s | (samples at: s) x@ (((newProposedS at: (s - windowFirst + 1)) + base))]
! !

!MUIEnvelopeEditorMorph methodsFor: 'events' stamp: 'yo 9/6/2016 16:37'!
commitEdit
	selectedSamples := nil.
	newSamples := nil.
	selectedControlPoints := nil.
	selectionInterval := nil.
	selection := nil.
	curve delete.
	curve := editingCurve.
! !

!MUIEnvelopeEditorMorph methodsFor: 'accessing' stamp: 'yo 9/6/2016 14:42'!
curve: aMorph
	curve := aMorph.
	self addMorph: curve.
	curve showOrHideHandles.
	self extent: curve extent + (50@50).
	! !

!MUIEnvelopeEditorMorph methodsFor: 'events' stamp: 'yo 8/25/2016 15:17'!
handlesMouseDown: evt
	^ true.
! !

!MUIEnvelopeEditorMorph methodsFor: 'events' stamp: 'yo 8/25/2016 15:17'!
handlesMouseMove: evt
	^ isSelecting
! !

!MUIEnvelopeEditorMorph methodsFor: 'events' stamp: 'yo 8/25/2016 15:17'!
handlesMouseUp: evt
	^ isSelecting! !

!MUIEnvelopeEditorMorph methodsFor: 'initialize' stamp: 'yo 9/6/2016 14:41'!
initialize
	super initialize.
	self color: Color white.
	isSelecting := false.
! !

!MUIEnvelopeEditorMorph methodsFor: 'events' stamp: 'yo 8/25/2016 15:13'!
mouseDown: evt

	| local |
	local := evt cursorPoint - self position.
	isSelecting := true.
	selectionStart := local.
! !

!MUIEnvelopeEditorMorph methodsFor: 'events' stamp: 'yo 8/25/2016 16:11'!
mouseMove: evt
	| local rect |
	isSelecting ifTrue: [
		local := evt cursorPoint - self position.
		rect := selectionStart corner: local.
		selectionPos := rect origin.
		selectionRect := (rect origin x@(self height - rect bottom)) extent: rect extent.
	].
! !

!MUIEnvelopeEditorMorph methodsFor: 'events' stamp: 'yo 9/6/2016 16:45'!
mouseUp: evt
	| newVertices base next step stepY p vs |
	isSelecting := false.
	selectionInterval := selectionRect left to: selectionRect right.
	selectedControlPoints := OrderedCollection new.
	newVertices := OrderedCollection new.
	windowLeft := OrderedCollection new.
	windowRight := OrderedCollection new.

	vs := curve vertices.
	1 to: vs size do: [:i |
		(vs at: i) x < selectionInterval first ifTrue: [
			newVertices add: (curve vertices at: i).
			(vs at: i) x + 20 >= selectionInterval first ifTrue: [
				windowLeft add: {i. vs at: i}]
		] ifFalse: [
			(selectionInterval includes: (vs at: i) x) ifTrue: [
				base := curve vertices at: i.
				next := curve vertices at: i + 1.
				step := next x - base x.
				stepY := next y - base y.
				0 to: step - 1 do: [:j |
					p := base + ((step@stepY) * (j asFloat / step)).
					newVertices add: p.
					selectedControlPoints add: {newVertices size. p}]]
			ifFalse: [
				newVertices add: (curve vertices at: i).
				(vs at: i) x - 20 <= selectionInterval last ifTrue: [
					windowRight add: {newVertices size. vs at: i}].
			
			]]].
	editingCurve := curve copy.
	editingCurve setVertices: newVertices asArray.
	self addMorph: editingCurve.
	self halt.

	selection := MUISelection new.
	selection extent: selectionRect extent.
	selection position: selectionPos + self position.
	selection color: (Color blue alpha: 0.5).
	selection editor: self.
	self addMorph: selection.
! !

!MUIEnvelopeEditorMorph methodsFor: 'events' stamp: 'yo 9/6/2016 16:47'!
selectionMoveBy: diff
	| v |
	v := editingCurve vertices copy.
	selectedControlPoints do: [:pair |
		v at: pair first put: pair second - (0@diff y).].
	editingCurve setVertices: v.
	windowLeft do: [:pair |
		v at: pair first put: pair second - (0@(diff y / 2.0))].
	windowRight do: [:pair |
		v at: pair first put: pair second - (0@(diff y / 2.0))].
	self changed.
! !

Morph subclass: #MUISelection
	instanceVariableNames: 'clickPoint origPos editor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MUI'!

!MUISelection methodsFor: 'accessing' stamp: 'yo 8/25/2016 16:00'!
editor: anEditor
	editor := anEditor
! !

!MUISelection methodsFor: 'events' stamp: 'yo 8/25/2016 15:59'!
handlesMouseDown: evt
	^ true.
! !

!MUISelection methodsFor: 'events' stamp: 'yo 8/25/2016 16:14'!
handlesMouseMove: evt
	^ clickPoint notNil

! !

!MUISelection methodsFor: 'events' stamp: 'yo 8/25/2016 16:14'!
handlesMouseUp: evt
	^ true
! !

!MUISelection methodsFor: 'events' stamp: 'yo 8/25/2016 16:17'!
mouseDown: evt
	clickPoint := evt cursorPoint.
	origPos := self position.
! !

!MUISelection methodsFor: 'events' stamp: 'yo 8/29/2016 14:57'!
mouseMove: evt
	| diff |
	diff := evt cursorPoint - clickPoint.
	self position: origPos x@(origPos y + diff y).
	editor selectionMoveBy: diff negated.
! !

!MUISelection methodsFor: 'events' stamp: 'yo 8/29/2016 15:22'!
mouseUp: evt
	clickPoint := nil.
	editor commitEdit.
	self delete.! !

Object subclass: #MUISmoother
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MUI'!

!MUISmoother class methodsFor: 'as yet unclassified' stamp: 'yo 9/15/2016 08:42'!
example1
	| c f points rect samples t |
	t := MUISmoother new.
	f := Form extent: 400@200 depth: 32.
	f fillColor: Color white.
	c := f getCanvas.
	samples := MUISmoother sample3 * 200.
	2 to: samples size do: [:i |
		c line: (i-1)@(samples at: (i-1)) to: i@(samples at: i) width: 1 color: Color black.
	].

	points := t process: samples * 5.0. 

	points do: [:p |
		rect := (p-1)@(samples at: p) extent: (2@2).
		c fillRectangle: ((p-2)@((samples at: p) - 2) extent: (5@5)) color: Color red].

	f display! !

!MUISmoother class methodsFor: 'as yet unclassified' stamp: 'yo 9/14/2016 10:51'!
sample1

	^ {0. 0.0027574901729717. 0.0025389544592485. 0.002503514389964. 0.0025687999166765. 0.0026611516431708. 0.0028233898747263. 0.35836517816109. 0.68773707227755. 0.75009932636513. 0.59489491616527. 0.42652257030049. 0.38199904916173. 0.58421364472953. 0.64785503303353. 0.5778711874517. 0.4285131489535. 0.33048032536804. 0.46034793179019. 0.52225720849032. 0.47302292281349. 0.33789254398583. 0.23450092188705. 0.54474133812188. 0.94307324695794. 1.0. 0.87748771120736. 0.61921030168267. 0.45095512633294. 0.42396434738845. 0.4549855371344. 0.36846649196394. 0.26774602985873. 0.18746255914972. 0.27209815884789. 0.37541107549969. 0.36007352791226. 0.26024739001995. 0.18081843559245. 0.14824914065851. 0.31756136566742. 0.38312664466617. 0.26472656753574. 0.18291622708757. 0.12638833859311. 0.087329747691003. 0.060341699993955. 0.041693877667609. 0.028808924842626. 0.019905920247393. 0.013754270476071. 0.0095036848567898. 0.0065666948262071. 0.0045373491742574. 0.0031351416765189. 0.0021662677447439. 0.0014968116142906. 0.0010342431207217. 0.0007146235125911. 0.0004937782410985. 0.00034118285999184. 0.00023574501754178. 0.0001628910703747. 0.00011255162178571. 7.7769058979621e-05. 5.3735615191891e-05. 3.7129295146917e-05. 2.5654963586547e-05. 1.772664577024e-05. 1.2248468394498e-05. 8.4632447572358e-06. 5.847787913908e-06. 4.0406095395614e-06. 2.7919129412158e-06. 1.9291091627339e-06. 1.3329447211399e-06. 9.2101572260053e-07. 6.3638805547857e-07. 4.3972080226593e-07. 3.0383101266047e-07. 2.099359166657e-07. 1.4505802602508e-07. 1.0022980888904e-07. 6.9255069230618e-08. 4.7852715514968e-08. 3.3064458103206e-08. 2.2846334807901e-08. 1.5785976619016e-08. 1.0907533128769e-08. 7.5367052743463e-09. 5.2075848626412e-09. 3.598250480986e-09. 2.4862598437886e-09. 1.7179142598253e-09. 1.1870148978658e-09. 8.2018367722679e-10. 5.6671706732223e-10. 3.9158072909795e-10. 2.705677618114e-10. 1.869525019958e-10. 1.2917724419223e-10. 8.9256733038453e-11. 6.1673128388655e-11. 4.2613891917629e-11. 2.9444640163138e-11. 2.0345150305491e-11. 1.4057752547613e-11. 9.7133939291566e-12. 6.7116028236186e-12. 4.6374678433058e-12. 3.2043184341192e-12. 2.2140675987228e-12. 1.5298405905019e-12. 1.0570628086674e-12. 7.3039168132595e-13. 5.0467387699461e-13. 3.4871123718706e-13. 2.409464902626e-13. 1.6648532480327e-13. 1.1503536911954e-13. 7.9485146144699e-14. 5.492128292061e-14. 3.7948614373945e-14. 2.6221136792058e-14. 1.8117826991912e-14. 1.2518742274796e-14. 8.649997923736e-15. 5.9768376890385e-15. 4.1297709928857e-15. 2.8535178349054e-15. 1.9716771518114e-15. 1.3623567169436e-15. 9.4133735997557e-16. 6.5042944392479e-16. 4.4942333805609e-16. 3.1053492276163e-16. 2.1456806004643e-16. 1.482586983659e-16. 1.0244140629298e-16. 7.0783278676362e-17. 4.890862460724e-17. 3.3794090507772e-17. 2.3350481548213e-17. 1.6134334915376e-17. 1.1148234422316e-17. 7.7030186563892e-18. 5.3225018862551e-18. 3.6776552823709e-18. 2.5411256687718e-18. 1.7558236442427e-18. 1.213209673933e-18. 8.3828415612677e-19. 5.7922341320622e-19. 4.0022178053697e-19. 2.7653850459631e-19. 1.9107796309835e-19. 1.3202780227936e-19. 9.122640401255e-20. 6.3034091741756e-20. 4.3554239626196e-20. 3.0094371649446e-20. 2.0794113385196e-20. 1.4367965239311e-20. 9.9277221407158e-21. 6.8596857545207e-21. 4.7397871536707e-21. 3.275014108675e-21. 2.2629092492692e-21. 1.5635812499529e-21. 1.0803725844235e-21. 7.464927591114e-22. 5.157940220431e-22. 3.5638961578791e-22. 2.4624700899611e-22. 1.7014252001134e-22. 1.1755717992646e-22. 8.1222609458808e-23. 5.6116766733438e-23. 3.8769512687925e-23. 2.6783200215661e-23. 1.8501122101991e-23. 1.2778511357121e-23. 8.824381516971e-24. 6.0922242548921e-24. 4.2044096575743e-24. 2.8999985006702e-24. 1.9986941975135e-24. 1.3759276551002e-24. 9.456202905887e-25. 6.4829277417949e-25. 4.4285037560836e-25. 3.0089726056467e-25. 2.028130844882e-25. 1.3504016916182e-25. 8.8211604198323e-26. 5.585490745214e-26. 3.3497603324774e-26. 1.8049477334367e-26. 7.375378324485e-27. 0.0 }! !

!MUISmoother class methodsFor: 'as yet unclassified' stamp: 'yo 9/14/2016 11:11'!
sample2

	^{ 0. 0.0013093833414529. 0.0019998560663388. 0.002563436568088. 0.0026840066952926. 0.0026988169229341. 0.0028626634629455. 0.0030532080772522. 0.0036761758325295. 0.0059879266563646. 0.015436113619389. 0.035045172333063. 0.072189223257952. 0.10994452778377. 0.18110984803291. 0.24196524358552. 0.31248945179122. 0.36594232835173. 0.44412194642407. 0.52585730886769. 0.59922182572922. 0.66930193438488. 0.7312798313171. 0.77364160157042. 0.80697193507028. 0.8418839369072. 0.87093471726379. 0.89247668789561. 0.90168385647672. 0.93773893407428. 0.97246103744505. 0.98700427922559. 1.0. 0.97595929638155. 0.94369006867053. 0.8850329791974. 0.84556398738956. 0.79676155557552. 0.74197769057425. 0.69559914603831. 0.64806147567731. 0.60123168685604. 0.55879064893768. 0.51556323975136. 0.47906940580123. 0.45146824924669. 0.42501139677405. 0.40410393595893. 0.3786713805815. 0.35746871322457. 0.34169149991046. 0.32205975307959. 0.30005705961674. 0.27956497610629. 0.25265726831671. 0.23368457725563. 0.22767859811531. 0.21929583781705. 0.21213570755276. 0.20205886705052. 0.18579585267973. 0.15604001592232. 0.12024206690764. 0.086378713634833. 0.060934647863634. 0.043166581694132. 0.030827751773307. 0.022062228658717. 0.015651028895111. 0.010814290037684. 0.0074722786829186. 0.0051630673493048. 0.0035674898734125. 0.0024650067715815. 0.0017032289582018. 0.0011768681898799. 0.00081317266198671. 0.00056187276734699. 0.00038823364404265. 0.00026825500630221. 0.00018535423728149. 0.00012807305290263. 8.8493789113843e-05. 6.11459530664e-05. 4.2249603011048e-05. 2.9192944407456e-05. 2.0171239411004e-05. 1.3937575262026e-05. 9.6303602700886e-06. 6.6542333550772e-06. 4.5978248503272e-06. 3.1769264649069e-06. 2.1951419327194e-06. 1.5167643024613e-06. 1.0480274636817e-06. 7.2414783797131e-07. 5.003599041606e-07. 3.4573040613704e-07. 2.3888695123874e-07. 1.650619553599e-07. 1.1405181727889e-07. 7.8805540039011e-08. 5.4451697107203e-08. 3.7624134638044e-08. 2.5996918355456e-08. 1.7962919518755e-08. 1.2411711547121e-08. 8.5760403809807e-09. 5.9257324090635e-09. 4.0944634280346e-09. 2.8291253009164e-09. 1.9548217473661e-09. 1.3507104314593e-09. 9.3329118295111e-10. 6.4487019494942e-10. 4.4558145177845e-10. 3.07880261565e-10. 2.127339727978e-10. 1.4699143068603e-10. 1.0156561030687e-10. 7.0178102568395e-11. 4.8490498098432e-11. 3.3505148278151e-11. 2.315082649609e-11. 1.5996372681252e-11. 1.1052908004184e-11. 7.6371521144071e-12. 5.2769943311381e-12. 3.6462081931701e-12. 2.5193959870839e-12. 1.7408099377408e-12. 1.202836239422e-12. 8.3111601058597e-13. 5.7427017963006e-13. 3.9679950190156e-13. 2.741740077986e-13. 1.8944414178549e-13. 1.3089892388615e-13. 9.0446349734731e-14. 6.2495145023721e-14. 4.3181850153767e-14. 2.9837068437705e-14. 2.0616321949012e-14. 1.424511983715e-14. 9.8428521668175e-15. 6.8010465687035e-15. 4.6992755199864e-15. 3.2470286735084e-15. 2.2435770019424e-15. 1.5502281614083e-15. 1.0711514121205e-15. 7.4012715220212e-16. 5.1140031108354e-16. 3.5335873292396e-16. 2.4415813864036e-16. 1.687044526766e-16. 1.1656845321149e-16. 8.0544502854893e-17. 5.5653326224079e-17. 3.8454395269623e-17. 2.6570548872631e-17. 1.8359284056221e-17. 1.2685593629504e-17. 8.7652713008539e-18. 6.0564694157811e-18. 4.1847957350022e-18. 2.8915390273696e-18. 1.9979411322067e-18. 1.3804976044743e-18. 9.5386914322885e-19. 6.5908471415945e-19. 4.5539863508079e-19. 3.1465891307634e-19. 2.1741342474757e-19. 1.5022026352392e-19. 1.037921788613e-19. 7.171213137733e-20. 4.9545975411504e-20. 3.422995900138e-20. 2.3647141272402e-20. 1.6334817264286e-20. 1.128225417718e-20. 7.7911257857894e-21. 5.3788840922129e-21. 3.7121119685579e-21. 2.5604321711338e-21. 1.7646662031135e-21. 1.2148210655001e-21. 8.3489671101097e-22. 5.7238354188579e-22. 3.9099650758187e-22. 2.6566474757378e-22. 1.7906521281558e-22. 1.1922811937899e-22. 7.7882875107301e-23. 4.9314769996422e-23. 2.9575277902291e-23. 1.5936033122821e-23. 6.5118072611542e-24. 0.0 }! !

!MUISmoother class methodsFor: 'as yet unclassified' stamp: 'yo 9/14/2016 11:11'!
sample3

	^{ 0. 0.0030963050862845. 0.0031041331574785. 0.0031405295212529. 0.0029557478125638. 0.0028316793354051. 0.044564789374236. 0.44094831344495. 0.66251291215019. 0.76153410192469. 0.80329073905198. 0.7597099369446. 0.6981608071748. 0.63948245742709. 0.6032408855863. 0.57395956535164. 0.58072717331875. 0.56871910444301. 0.54387433233797. 0.51648862345965. 0.50183524352803. 0.49596366441437. 0.49442425914775. 0.55210564216085. 0.72027461463579. 1.0. 0.80772195306594. 0.55918432139278. 0.38637574536573. 0.2669715660861. 0.18446756806383. 0.12746032238574. 0.088070365704919. 0.060853412346631. 0.042047481066454. 0.029053273967361. 0.020074735961274. 0.013870912605644. 0.0095842871056986. 0.0066223903544315. 0.0045758261692484. 0.0031617305474034. 0.0021846395243738. 0.0015095052482697. 0.0010430126506695. 0.00072068392546933. 0.00049796611395633. 0.0003440762307398. 0.00023774400373593. 0.0001642724281252. 0.0001135062559576. 7.842861399119e-05. 5.4191297666366e-05. 3.7444190607691e-05. 2.5872569057524e-05. 1.7876981870702e-05. 1.2352344404543e-05. 8.5350269379152e-06. 5.8973861025559e-06. 4.0748750859962e-06. 2.8155918490792e-06. 1.9454728313202e-06. 1.3442489088416e-06. 9.2882613559188e-07. 6.4178523857465e-07. 4.4345048150194e-07. 3.0640775213932e-07. 2.1171645248912e-07. 1.4628833174488e-07. 1.0107985551023e-07. 6.9842397360989e-08. 4.8258557021655e-08. 3.3344919815668e-08. 2.3040100639939e-08. 1.5919845469066e-08. 1.1000030507428e-08. 7.6006252018792e-09. 5.2517514777156e-09. 3.6287643769227e-09. 2.5073442611722e-09. 1.7324840930889e-09. 1.19708326215e-09. 8.2714053621706e-10. 5.7152406021955e-10. 3.9490222576927e-10. 2.7286299219708e-10. 1.8853837398752e-10. 1.3027307008042e-10. 9.001390525389e-11. 6.2196318431486e-11. 4.2975377442175e-11. 2.9694399465894e-11. 2.0517739668389e-11. 1.4177017832174e-11. 9.7957928087549e-12. 6.7685288012237e-12. 4.6768050871648e-12. 3.2315006155074e-12. 2.2328487374747e-12. 1.5428173518286e-12. 1.0660299449804e-12. 7.3658741200041e-13. 5.089547024589e-13. 3.5166922597165e-13. 2.4299056740717e-13. 1.6789743130194e-13. 1.1601089090503e-13. 8.0159280730822e-14. 5.5387127574258e-14. 3.8270471680163e-14. 2.6443489858758e-14. 1.8271482720968e-14. 1.2624922471836e-14. 8.7233620381439e-15. 6.0275238735832e-15. 4.1647982820897e-15. 2.8777204196533e-15. 1.9883985872099e-15. 1.3739114505308e-15. 9.4932266669813e-16. 6.5594696387204e-16. 4.5323514516778e-16. 3.1316899739025e-16. 2.1638847007455e-16. 1.4951639715815e-16. 1.0331025683438e-16. 7.1383627694532e-17. 4.9323460423722e-17. 3.4080698029269e-17. 2.3548515693718e-17. 1.6271174749643e-17. 1.1242775915459e-17. 7.7683407684701e-18. 5.3676430743351e-18. 3.708849350845e-18. 2.5626764800473e-18. 1.7707142141087e-18. 1.2234993248762e-18. 8.4539406999613e-19. 5.841356645132e-19. 4.0361583652564e-19. 2.7888386828684e-19. 1.9269843876722e-19. 1.3314756757104e-19. 9.2000002196864e-20. 6.356867402741e-20. 4.3923630036029e-20. 3.0349601943413e-20. 2.0970465515915e-20. 1.4489830779061e-20. 1.0011943080884e-20. 6.9178858819763e-21. 4.7800072782714e-21. 3.3028084129137e-21. 2.2821202308119e-21. 1.5768616036652e-21. 1.0895528897869e-21. 7.5284083450953e-22. 5.2018535840052e-22. 3.5942913453195e-22. 2.4835212022827e-22. 1.716021001829e-22. 1.1857073303353e-22. 8.192796941134e-23. 5.6609183193613e-23. 3.9114827735078e-23. 2.7026866592801e-23. 1.8674525210392e-23. 1.2903369985391e-23. 8.9157148501172e-24. 6.1603917295233e-24. 4.2565603590159e-24. 2.9410864499273e-24. 2.0321402340052e-24. 1.404091794819e-24. 9.7013339938075e-25. 6.7028506234249e-25. 4.6310027119857e-25. 3.1994311443879e-25. 2.2102686283078e-25. 1.5267948192025e-25. 1.054538622481e-25. 7.2822676278928e-26. 5.0275733638076e-26. 3.4696640911416e-26. 2.3932058399872e-26. 1.6494108736343e-26. 1.1354768486605e-26. 7.8036709863092e-27. 5.3499968148417e-27. 3.654593272568e-27. 2.483135534105e-27. 1.6737022785579e-27. 1.1144104463851e-27. 7.2796075173574e-28. 4.609392695101e-28. 2.7643709653816e-28. 1.4895211574094e-28. 6.0864863006042e-29. 0.0 }! !

!MUISmoother methodsFor: 'all' stamp: 'yo 9/15/2016 08:40'!
errorBetween: startIndex and: endIndex in: array
	| n a aa ab ac ax ay b yk diff error |
	n := endIndex - startIndex + 1.
	aa := 0.
	ax := 0.
	ay := 0.
	ab := 0.
	ac := 0.
	startIndex to: endIndex do: [:xk |
		yk := array at: xk.
		aa := aa + (xk * yk).
		ax := ax + xk.
		ay := ay + yk.
		ab := ab + (xk * xk).
		ac := ac + xk.
		].
	a := ((n asFloat * aa) - (ax * ay)) / ((n * ab) - (ac * ac)).
	b := (ab asFloat * ay) - (aa * ax) / ((n * ab) - (ac * ac)).

	error := 0.
	startIndex to: endIndex do: [:xk |
		diff := (a * xk + b) - (array at: xk).
		error := error + (diff * diff)].
	^ error! !

!MUISmoother methodsFor: 'all' stamp: 'yo 9/14/2016 15:42'!
isPeakAt: index in:  anArray window: window
	| leftScore rightScore peak half |
	
	half := window // 2 + 1.

	leftScore := 0.
	rightScore := 0.
	window to: 1 by: -1 do: [:i |
		(anArray at: (index - i)) < (anArray at: index) ifTrue: [leftScore := leftScore + 1].
		(anArray at: index) > (anArray at: (index + i)) ifTrue: [rightScore := rightScore + 1]].
	(leftScore >= half and: [rightScore >= half]) ifTrue: [
		peak := true.
		index - half to: index + half do: [:i | (anArray at: index) < (anArray at: i) ifTrue: [peak := false]].
		^ peak].
	^ false.
! !

!MUISmoother methodsFor: 'all' stamp: 'yo 9/14/2016 15:44'!
isValleyAt: index in: anArray window: window
	| leftScore rightScore valley half |
	
	half := window // 2 + 1.

	leftScore := 0.
	rightScore := 0.
	window to: 1 by: -1 do: [:i |
		(anArray at: (index - i)) > (anArray at: index) ifTrue: [leftScore := leftScore + 1].
		(anArray at: index) < (anArray at: (index + i)) ifTrue: [rightScore := rightScore + 1]].
	(leftScore >= half and: [rightScore >= half]) ifTrue: [
		valley := true.
		index - half to: index + half do: [:i | (anArray at: index) > (anArray at: i) ifTrue: [valley := false]].
		^ valley].
	^ false.
! !

!MUISmoother methodsFor: 'old tests' stamp: 'yo 9/13/2016 22:21'!
last: n monotonicIn: aCollection
	| e |
	e := nil.
	aCollection size - n + 1 to: aCollection size do: [:i |
		e ifNotNil: [(aCollection at: i) > e ifFalse: [^ false]].
		[e := aCollection at: i]].
	^ true.
! !

!MUISmoother methodsFor: 'old tests' stamp: 'yo 9/13/2016 22:18'!
last: n sameDirAs: generalDir in: aCollection
	 aCollection size - n + 1 to: aCollection size do: [:i |
		(aCollection at: i) sign ~= generalDir ifTrue: [^ false]].
	^ true.
! !

!MUISmoother methodsFor: 'old tests' stamp: 'yo 9/15/2016 08:42'!
oldProcess: anArray
	| start result lastError end error lastSlopes lastErrors keepGoing generalDir newEnd maybeDir step |
	step := 1.
	start := 1.
	anArray size <= step ifTrue: [^ {1. anArray size}].
	result := OrderedCollection new.
	lastError := 0.
	[true] whileTrue: [
		end := start + step.
		end > anArray size ifTrue: [^ result add: anArray size; asArray].
		error := self errorBetween: start and: end in: anArray.
		lastSlopes := OrderedCollection with: ((anArray at: end) - (anArray at: start)).
		lastErrors := OrderedCollection with: error.
		keepGoing := true.
		generalDir := nil.
		[keepGoing] whileTrue: [
			newEnd := end + step.
			newEnd > anArray size ifTrue: [^ result add: start; add: anArray size; asArray].
			lastSlopes add: (anArray at: newEnd) - (anArray at: end).
			lastErrors add: (self errorBetween: start and: newEnd in: anArray).
			(generalDir isNil and: [lastSlopes size >= 3]) ifTrue: [
				maybeDir := (lastSlopes at: (lastSlopes size - 2)) sign.
				(self last: 2 sameDirAs: maybeDir sign in: lastSlopes) ifTrue: [
					generalDir := maybeDir].
			].
			generalDir notNil ifTrue: [
				((self last: 3 sameDirAs: (generalDir * -1) in: lastSlopes)
				 	and: [(self last: 3 monotonicIn: lastErrors)]) ifTrue: [
					"direction changed"
					result add: start.
					start := newEnd - (step * 3).
					keepGoing := false]
				ifFalse: [
					(lastErrors size >= 6 and: [
						(lastErrors last: 3) sum > ((lastErrors at: lastErrors size - 3) * lastErrors size)]) ifTrue: [
						"direction changed"
						result add: start.
						start := newEnd - (step * 3).
						keepGoing := false]
					ifFalse: [end := newEnd]]]]].
	^ result asArray
		
					! !

!MUISmoother methodsFor: 'all' stamp: 'yo 9/15/2016 08:42'!
process: anArray
	| interval residuals maybeThere start  e error index min result s peak peaks bigErrors close peaksAndValleys valleys mid |

	interval := 5.
	anArray size <=(interval * 2)  ifTrue: [^ {1. anArray size}].
	
	residuals := (1 to: anArray size - (interval - 1)) collect: [:i |
		self errorBetween: i and: i + (interval - 1) in: anArray].
	start := interval + 1.

	bigErrors := OrderedCollection new.
	
	[start < (residuals size - interval)] whileTrue: [
		peak := self isPeakAt: start in: residuals window: 5.
		peak ifTrue: [
			bigErrors add: start.
			start  := start + interval]
		ifFalse: [start := start + 1]].
	
	peaks := OrderedCollection new.
	interval + 1 to: (anArray size - interval) do: [:p |
		(self isPeakAt: p in: anArray window: 5) ifTrue: [
			peaks add: p]].

	valleys := OrderedCollection new.
	interval + 1 to: (anArray size - interval) do: [:p |
		(self isValleyAt: p in: anArray window: 5) ifTrue: [
			valleys add: p]].
	
	peaksAndValleys := SortedCollection new.
	peaksAndValleys addAll: peaks.
	peaksAndValleys addAll: valleys.

	maybeThere := SortedCollection withAll: peaksAndValleys.
	maybeThere add: 1.
	maybeThere add: anArray size.

	bigErrors do: [:m |
		close := peaksAndValleys detect: [:p | (m - p) abs < (interval // 2 + 1)] ifNone: [nil].
		close ifNil: [
			maybeThere add: m]].

	result := OrderedCollection new.
	1 to: maybeThere size - 2 do: [:i |
		s := maybeThere at: i.
		e := maybeThere at: (i + 2).
		mid := maybeThere at: i + 1.
		(peaksAndValleys includes: mid) ifTrue: [
			result add: mid]
		ifFalse: [
			min := Float infinity.
			index := 0.
			s + 1 to: (e -1) do: [:n |
				error := ((self errorBetween: s and: n in: anArray)+ (self errorBetween: n and: e in: anArray)).
				error < min ifTrue: [
					min := error.
					index := n]].
			result add: index]].
	^ result asArray
			! !
